
import { createStore, applyMiddleware, compose } from 'redux';
import isArray from 'lodash/isArray';
import getDisplayName from 'react-display-name';
import URI from 'urijs';

// middlewares
import { createLogger } from 'redux-logger';
import promisify from 'redux-promise';
import thunk from 'redux-thunk';

// for getData
import { matchPath } from 'react-router';

import history from 'roderic/libs/history';

import { setJwtToken } from 'transport';

import { routerMiddleware } from 'react-router-redux';

import tlog from 'roderic/webpack/logt'

import log from 'roderic/webpack/logn';

const wrapDispatchWithMiddlewares = (store, middlewares) => {
    middlewares.forEach(middleware => store.dispatch = middleware(store)(store.dispatch))
};

const node = typeof global !== 'undefined' && Object.prototype.toString.call(global.process) === '[object process]';

const triggerMultiple = store => next => action => {

    if (isArray(action.type)) {

        return action.type.forEach(a => next({
            type: a
        }));
    }

    return next(action);
};

let resolve;

/**
 * import { getStore } from 'roderic/libs/configureStore';
 * getStore.then(store => log('get colors: ', getSearchListInitColors(store.getState())));
 *
 * @type {Promise<any>}
 *
 * Use only in browser
 */
export const getStore = new Promise(res => resolve = res);

const configureStore = opt => {

    const {
        preloadedState,
        reducers,
        createLoggerConfig = {},
        composeEnhancersConfig = {}, // https://github.com/zalmoxisus/redux-devtools-extension/blob/master/docs/API/Arguments.md#actionsblacklist--actionswhitelist
    } = opt;

    const middlewares = [triggerMultiple, thunk, promisify];

    // http://extension.remotedev.io/#usage
    let composeEnhancers = compose;

    if ( ! node ) {

        const ls = (localStorage && localStorage.getItem && localStorage.getItem('debug'));

        log(`localStorage.getItem('debug')`, ls);

        if ( (process && eval('process.env') && process.env.NODE_ENV !== 'production') || ls ) {

            middlewares.push(createLogger(createLoggerConfig));

            composeEnhancers = window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__
                ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__(composeEnhancersConfig)
                : compose;
        }

        middlewares.push(routerMiddleware(history))
    }

    let store;

    if (preloadedState) {

        store = createStore(
            reducers,
            preloadedState,
            composeEnhancers(
                applyMiddleware(...middlewares) // applyMiddleware returns an redux enhancer
            )
        );
    }
    else {

        store = createStore(
            reducers,
            composeEnhancers(
                applyMiddleware(...middlewares) // applyMiddleware returns an redux enhancer
            )
        );
    }

    resolve(store);

    return store;
};

export default configureStore;

const buildUrls = (routes, prefix = '') => {

    if ( isArray(routes) ) {

        return routes.map(route => {

            const node = {...route};

            node.path = `${prefix}${node.path}`;

            if (node.nested) {

                if ( ! isArray(node.nested) ) {

                    throw `'node.nested' is not array`;
                }

                node.nested = buildUrls(node.nested, node.path);
            }

            return node;
        });
    }

    throw `buildUrls: routes parameter is mandatory`;
};

const extract = (routes, url, rootTrigger = [], root) => {

    let list = [];

    let notRoot = !root;

    const found = routes.find(route => matchPath(url, route));

    if (found) {

        if ( notRoot ) {

            // take out component, nested, trigger
            const { component, nested, trigger, ...rest} = found;

            root = rest;
        }

        list.push({...found});

        if (found.nested) {

            if ( ! isArray(found.nested) ) {

                throw `'found.nested' exist but is not array`;
            }

            list = list.concat(extract(found.nested, url, rootTrigger, root))
        }

        if (found.trigger) {

            list = list.concat(found.trigger.map(component => ({
                ...root,
                component,
            })));
        }
    }

    if (found && notRoot && rootTrigger && rootTrigger.length) {

        list = list.concat(rootTrigger.map(component => ({
            ...root,
            component,
        })));
    }

    return list;
};

/**
 * Stop using after implementing real metatags
 */
export const cascadeMeta = (routesObj, parentMeta) => {

    if ( isArray(routesObj.nested) ) {

        routesObj.nested.forEach(route => {

            if ( ! route.meta ) {

                route.meta = parentMeta || routesObj.meta;
            }

            if (route.nested) {

                if ( ! isArray(route.nested) ) {

                    throw `cascadeMeta: 'node.nested' is not array`;
                }

                cascadeMeta(route, route.meta);
            }
        });
    }
};

export const fetchData = opt => {

    const {
        logFlag = 'front'
    } = opt;

    let {
        routes,
        trigger,
    } = opt;

    routes = buildUrls(routes);

    return opt => {

        const {
            url,
            store,
            jwtToken,
        } = opt;

        jwtToken && setJwtToken(jwtToken);

        const urlParts = url.split('?');

        const getParameters = URI.parseQuery(urlParts[1]);

        const routesList = extract(routes, urlParts[0], trigger);

        let firstRoute = false;

        if (routesList.length) {

            firstRoute = routesList[0];
        }

        return Promise.all(routesList.map(route => {

            let
                promise,
                foundComponent  = !!(route && route.component),
                componentName   = foundComponent ? getDisplayName(route.component) : '<displayName>',
                isFetchData     = (foundComponent && typeof route.component.fetchData === 'function')
            ;

            if ( foundComponent ) {

                if ( isFetchData ) {

                    try {

                        promise = route.component.fetchData(store, matchPath(urlParts[0], route), getParameters).catch(e => {
                            tlog(`${logFlag}: component: ${componentName}, route: ${url}, generated rejected promise with data`);
                            log.dump(e)
                        });
                    }
                    catch (e) {

                        const reason = {
                            message: `${logFlag}: executing fetchData crashed in component: ${componentName}, route: ${url}, reason: `,
                            exception: e
                        }

                        tlog(reason.message, reason.exception);

                        return Promise.reject(reason);
                    }
                }
                else {

                    tlog(`${logFlag}: fetchData not found in component: ${componentName}, route: ${url}`);
                }
            }
            else {

                tlog(`${logFlag}: component behind router not found, url: ${url}`);
            }

            tlog(`${logFlag}: in component: ${componentName}, route: ${url}, promise: `, promise)

            if (isFetchData && ( ! promise || typeof promise.then !== 'function') ) {

                tlog(`${logFlag}: fetchData should return promise in component: ${componentName}, route: ${url}`, ' returned: ', promise);
            }

            return Promise.resolve(promise);
        }))
            .then(() => firstRoute)
        ;
    }
}



